import json
import logging
import os.path
import shutil
import sys
from time import sleep

from selenium.common.exceptions import WebDriverException
from selenium.webdriver import Chrome, Firefox
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager

from best_buy_bullet_bot.audio import sound_effects
from best_buy_bullet_bot.data import HEADLESS_WARNED, SHARED_DIR
from best_buy_bullet_bot.utils import (
    Colors,
    loading,
    print_table,
    validate_num,
    warnings_suppressed,
    yes_or_no,
)

SETTINGS_DIR = os.path.join(SHARED_DIR, "settings.json")
SOUND_MODES = ["disabled", "single", "repeat"]
DEFAULT_SETTINGS = {
    "funds": 1000,
    "tax": 0.095,
    "auto checkout": True,
    "account verification": True,
    "browser": "firefox",
    "sound mode": SOUND_MODES[2],
    "threads": 1,
}


def _save(data):
    with open(SETTINGS_DIR, "w+") as f:
        json.dump(data, f)


# Get the current settings
CURRENT_SETTINGS = DEFAULT_SETTINGS.copy()
if os.path.isfile(SETTINGS_DIR):
    save = False

    with open(SETTINGS_DIR) as f:
        for key, value in json.load(f).items():
            if key in CURRENT_SETTINGS:
                CURRENT_SETTINGS[key] = value
            elif not HEADLESS_WARNED():
                Colors.warn(
                    f"{key} is no longer supported and will be removed from your settings."
                )
                save = True

    if save and not HEADLESS_WARNED():
        if not warnings_suppressed() and yes_or_no(
            "Delete unsupported settings from your settings file (y/n): "
        ):
            _save(CURRENT_SETTINGS)
        HEADLESS_WARNED.update(True)
else:
    _save(DEFAULT_SETTINGS)


def get_settings():
    # A copy is returned so the settings can be safely manipulated
    return CURRENT_SETTINGS.copy()


def update_setting(setting, new_val):
    CURRENT_SETTINGS[setting] = new_val
    _save(CURRENT_SETTINGS)


def view_settings(show_default=False):
    settings = (DEFAULT_SETTINGS if show_default else CURRENT_SETTINGS).copy()

    settings["funds"] = f"${settings['funds']:,.2f}"
    settings["tax"] = f"{settings['tax'] * 100:.2f}%"
    settings["browser"] = settings["browser"].title()

    # Hidden property
    del settings["account verification"]

    rows = [[k.title(), v] for k, v in settings.items()]
    print_table(["Property", "Value"], rows)


def set_funds():
    while True:
        funds = input("Allotted money: $")
        funds = validate_num(funds.replace("$", ""), float)
        if funds is None or funds < 0:
            Colors.print(
                "Invalid input for funds. Please enter a positive number.",
                properties=["fail"],
            )
        else:
            break

    update_setting("funds", funds)
    Colors.print(f"Successfully set funds to ${funds:,.2f}!", properties=["success"])


def set_tax():
    while True:
        tax = input("Sales tax rate (%): ")
        tax = validate_num(tax.replace("%", ""), float)
        if tax is None or tax < 0:
            Colors.print(
                "Invalid input. Please enter a positive percentage for tax.",
                properties=["fail"],
            )
        else:
            break

    update_setting("tax", tax / 100)
    Colors.print(
        f"Successfully set the state sales tax rate to {tax:,.2f}%",
        properties=["success"],
    )


class MoneyManager:
    def get_funds(self):
        return CURRENT_SETTINGS["funds"]

    def check_funds(self, cost):
        return CURRENT_SETTINGS["funds"] - cost >= 0

    def make_purchase(self, cost):
        update_setting("funds", CURRENT_SETTINGS["funds"] - cost)


def _toggle_setting(setting_name):
    update_setting(setting_name, not CURRENT_SETTINGS[setting_name])
    Colors.print(
        f"Successfully {'enabled' if CURRENT_SETTINGS[setting_name] else 'disabled'} {setting_name}!",
        properties=["success"],
    )


def toggle_auto_checkout():
    _toggle_setting("auto checkout")


class DriverClassWrapper:
    def __init__(self, driver, manager, options):
        self.driver = driver
        self.manager = manager
        self.options = options


logging.disable(logging.WARNING)
DRIVER_NAMES = {
    "chrome": DriverClassWrapper(Chrome, ChromeDriverManager, ChromeOptions),
    "firefox": DriverClassWrapper(Firefox, GeckoDriverManager, FirefoxOptions),
}


def is_installed(browser_name):
    """Check if browser is installed

    Done by installing the drivers and trying to open the browser with selenium.
    If we can successfully open the browser then it is installed.
    """
    browser_name = browser_name.lower()

    if browser_name in DRIVER_NAMES:
        wrap = DRIVER_NAMES[browser_name]
    else:
        raise ValueError(
            f"3B Bot does not support {browser_name.title()}. Please pick either Chrome or Firefox."
        )

    # Install the drivers
    try:
        manager = wrap.manager()
        path = manager.install()
    except ValueError:
        return False

    options = wrap.options()
    options.add_argument("--headless")

    if CURRENT_SETTINGS["browser"] == "chrome":
        options.add_experimental_option("excludeSwitches", ["enable-logging"])

    try:
        # Try to open a browser window
        driver = wrap.driver(executable_path=path, options=options)
        driver.quit()
        return True

    except (WebDriverException, ValueError):
        # Delete the drivers we just installed
        name = manager.driver.get_name()
        driver_dir = path.split(name)[0] + name
        if os.path.isdir(driver_dir):
            shutil.rmtree(driver_dir)

        return False


def change_browser():
    with loading("Detecting browsers"):
        available_browsers = [
            browser.title() for browser in DRIVER_NAMES.keys() if is_installed(browser)
        ]

    if not len(available_browsers):
        Colors.print(
            "No available browsers. Please install either Chrome or Firefox and try again.",
            "Chrome can be installed from https://www.google.com/chrome/.",
            "Firefox can ben installed from https://www.mozilla.org/en-US/firefox/new/.",
            sep="\n",
            properties=["fail", "bold"],
        )
        sys.exit()

    # Print available browsers
    print("\n • ".join(["Available Browsers:"] + available_browsers), "\n")

    while True:
        new_browser = input("Select a browser from the list above: ").strip().title()

        if new_browser == "":
            continue

        if new_browser not in available_browsers:
            Colors.print("Invalid selection. Try again.", properties=["fail"])
        else:
            break

    update_setting("browser", new_browser.lower())
    Colors.print(
        f"Successfully changed browser to {new_browser}!", properties=["success"]
    )


def test_sound(repetitions=3, print_info=True):
    if print_info:
        print("Playing sound...")
        sleep(0.15)

    for i in range(repetitions):
        sound_effects.play(block=True)


def set_sound_mode():
    while True:
        sound_mode = input(
            f"Select a sound mode ({SOUND_MODES[0]}/{SOUND_MODES[1]}/{SOUND_MODES[2]}): "
        )
        if sound_mode not in SOUND_MODES:
            Colors.print(
                f'Invalid input for sound mode. Please enter either "{SOUND_MODES[0]}" (no sound),'
                f' "{SOUND_MODES[1]}" (plays sound once after coming back in stock), or "{SOUND_MODES[2]}"'
                " (plays sound repeatedly until item is no longer in stock).",
                properties=["fail"],
            )
        else:
            break

    update_setting("sound mode", sound_mode)
    Colors.print(
        f"Successfully set sound mode to {sound_mode}!", properties=["success"]
    )

    sleep(1)
    print("\nThat sounds like...")
    sleep(0.75)

    # No sound
    if sound_mode == SOUND_MODES[0]:
        sleep(0.5)
        print("Nothing. Crazy, right?")

    # Play sound once
    elif sound_mode == SOUND_MODES[1]:
        test_sound(repetitions=1, print_info=False)

    # Play sound on repeat
    elif sound_mode == SOUND_MODES[2]:
        test_sound(print_info=False)


def set_threads():
    while True:
        threads = input("Threads (per URL): ")
        threads = validate_num(threads, int)
        if threads is None or threads < 1:
            Colors.print(
                "Invalid number of threads. Please enter an integer greater than or equal to 1.",
                properties=["fail"],
            )
        else:
            break

    update_setting("threads", threads)
    Colors.print(
        f"Now using {threads} threads to track each URL!", properties=["success"]
    )


def reset_settings():
    print("Default settings:")
    view_settings(show_default=True)
    print()

    if yes_or_no("Reset (y/n): "):
        _save(DEFAULT_SETTINGS)
        Colors.print("Successfully reset settings!", properties=["success"])
    else:
        print("Settings reset aborted.")



import json
import os
from json.decoder import JSONDecodeError

from best_buy_bullet_bot.data import SHARED_DIR

COOKIES_DIR = os.path.join(SHARED_DIR, "login_cookies.json")


def save_cookies(driver):
    with open(COOKIES_DIR, "w+") as f:
        json.dump(driver.get_cookies(), f)


def load_cookies(driver):
    try:
        with open(COOKIES_DIR) as f:
            for cookie in json.load(f):
                driver.add_cookie(cookie)
        return True

    # An error occurred while adding the cookies
    except AssertionError:
        # Delete previous cookies
        delete_cookies()
        return False


def cookies_available():
    file_exists = os.path.isfile(COOKIES_DIR)

    if file_exists:
        # Make sure file is JSON decodable
        try:
            with open(COOKIES_DIR) as f:
                json.load(f)
            return True

        except JSONDecodeError:
            delete_cookies()

    return False


def delete_cookies():
    os.remove(COOKIES_DIR)



import json
import os
import sys
import tempfile
from glob import glob

from best_buy_bullet_bot.utils import Colors


def _read_file():
    FP.seek(0)
    return json.load(FP)


def _write_file(content):
    FP.seek(0)
    FP.write(json.dumps(content))
    FP.truncate()


# This function is called before killing all processes to
# make sure the temp file is deleted
def close_data():
    FP.close()
    if num_temp_files == 1 and os.path.isfile(FP.name):
        os.remove(FP.name)


# Windows doesn't allow for temporary files to be opened by a subprocess
# by default so we have to pass a flag to do so
def temp_opener(name, flag, mode=0o777):
    return os.open(name, flag | os.O_TEMPORARY, mode)


# TODO: Prevent directory from changing based on whether or not we are in root
temp_dir = os.path.dirname(tempfile.mkdtemp())

prefix = "best_buy_bullet_bot_global_vars"
suffix = ".json"
available_temp_files = glob(os.path.join(temp_dir, f"{prefix}*{suffix}"))
num_temp_files = len(available_temp_files)


if num_temp_files == 1:
    # Open the existing temp file
    FP = open(
        os.path.join(temp_dir, available_temp_files[0]),
        "r+",
        opener=temp_opener if sys.platform == "win32" else None,
    )
else:
    if num_temp_files > 1:
        # Too many temp files
        Colors.warn(
            f"Too many temporary files detected: {available_temp_files}. Deleting all temporary files."
        )
        for filename in available_temp_files:
            os.remove(os.path.join(temp_dir, filename))

    # Create a new temp file since we don't have any
    FP = tempfile.NamedTemporaryFile("r+", prefix=prefix, suffix=suffix, dir=temp_dir)
    _write_file({})


class ReferenceVar:
    """Points to a specific variable in the temp file.

    If a variale in changed by one process all other processes
    with that variable will receive that change when trying to
    access the variable.
    """

    def __init__(self, var_name):
        self.var_name = var_name

    def __new__(cls, var_name):
        # Return the value of the variable if it is a constant
        # else return this object
        self = super().__new__(cls)
        self.__init__(var_name)
        return self() if var_name.endswith("constant") else self

    def __call__(self):
        return _read_file()[self.var_name]

    def update(self, new_value, constant=False):
        # Update the value of the variable in the temp file
        updated_dict = _read_file()

        new_name = self.var_name
        new_name += "_constant" if constant else ""
        updated_dict.update({new_name: new_value})
        _write_file(updated_dict)

        return new_value if constant else self


if num_temp_files != 1:
    # We are in the main process. This is where variables are created.

    from keyring.util import platform_

    # We store data here so it doesn't get overwritten during an update
    shared_dir = os.path.join(
        os.path.dirname(platform_.data_root()), "best_buy_bullet_bot"
    )
    if not os.path.isdir(shared_dir):
        os.makedirs(shared_dir)

    # Save to temporary file
    HEADLESS_WARNED = ReferenceVar("HEADLESS_WARNED").update(False)
    SHARED_DIR = ReferenceVar("SHARED_DIR").update(shared_dir, constant=True)
else:
    # We are in a separate process. This is where variables are copied over from the main process.

    # Copy over all variables in the temp file to the locals dict so they can be imported
    for var_name in _read_file():
        locals()[var_name.replace("_constant", "")] = ReferenceVar(var_name)



import json
import os.path

from bs4 import BeautifulSoup
from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException
from requests.packages.urllib3.util.retry import Retry

from best_buy_bullet_bot.browser import get_user_agent
from best_buy_bullet_bot.data import SHARED_DIR
from best_buy_bullet_bot.utils import (
    Colors,
    loading,
    print_table,
    validate_num,
    yes_or_no,
)

URL_DIR = os.path.join(SHARED_DIR, "urls.json")


def _read():
    with open(URL_DIR) as f:
        return json.load(f)


def _save(data):
    with open(URL_DIR, "w+") as f:
        json.dump(data, f)


if not os.path.isfile(URL_DIR):
    _save({})


def get_url_data():
    items = _read().items()
    return [[item.split("\n"), qty] for item, qty in items]


def view_urls(show_qty=True):
    data = _read().items()

    columns = ["URL"] + (["Quantity"] if show_qty else [])
    rows = [[url, qty] for url, qty in data] if show_qty else zip(list(zip(*data))[0])
    print_table(columns, rows)


def get_qty():
    while True:
        qty = input("Quantity (optional): ")

        if qty.strip() == "" or qty == "inf":
            return "inf"

        qty = validate_num(qty, int)
        if qty is None or qty < 1:
            Colors.print(
                "Invalid input for quantity. Please enter an integer greater than or equal to 1.",
                properties=["fail"],
            )
        else:
            return qty


class QtyManager:
    def __init__(self, qty):
        self.qty = qty

    def get(self):
        return self.qty

    def decrement(self):
        self.qty -= 1


def add_url():
    new_url = input("URL to add: ")
    if new_url.strip() == "":
        print("Aborted.")
        return

    urls = _read()
    qty = get_qty()
    urls[new_url] = qty
    _save(urls)
    Colors.print(
        f"Successfully added {new_url}{'' if qty == 'inf' else f' with a quantity of {qty}'}!",
        properties=["success"],
    )


def add_url_group():
    url_group = []
    i = 1

    while True:
        new_url = input("URL to add: ")

        if new_url.strip() == "":
            if i == 1:
                print("Aborted.")
                break
            else:
                continue

        url_group.append(new_url)

        if i >= 2 and not yes_or_no("Would you like to add another URL (y/n): "):
            break

        i += 1

    urls = _read()
    qty = get_qty()
    urls["\n".join(url_group)] = qty
    _save(urls)
    Colors.print(
        f"Successfully added a URL group with {len(url_group)} URLs{'' if qty == 'inf' else f' and a quantity of {qty} for the group'}!",
        properties=["success"],
    )


def remove_url():
    urls = _read()
    ids = range(1, len(urls) + 1)
    rows = list(zip(ids, urls.keys()))
    print_table(["ID", "Active URLs"], rows, justifications=["center", "left"])

    while True:
        url_id = input("URL ID to remove: ").strip()

        if url_id == "":
            continue

        url_id = validate_num(url_id, int)
        if url_id is None or url_id < ids[0] or url_id > ids[-1]:
            Colors.print(
                f"Please enter valid URL ID between {ids[0]}-{ids[-1]}. Do not enter the URL itself.",
                properties=["fail"],
            )
        else:
            break

    selected_url = list(urls.keys())[url_id - 1]
    del urls[selected_url]
    _save(urls)

    comma_separated_selection = selected_url.replace("\n", ", ")
    Colors.print(
        f"Successfully removed: {comma_separated_selection}", properties=["success"]
    )


def get_url_titles():
    flattened_urls = [
        url
        for url_group, _ in get_url_data()
        for url in (url_group if type(url_group) is list else [url_group])
    ]

    session = Session()
    session.headers.update({"user-agent": get_user_agent()})
    retry = Retry(
        connect=3,
        backoff_factor=1,
        status_forcelist=[429, 500, 502, 503, 504],
        method_whitelist=["HEAD", "GET", "OPTIONS"],
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("https://", adapter)
    session.mount("http://", adapter)

    for url in flattened_urls:
        try:
            response = session.get(url, timeout=10)
        except RequestException as e:
            Colors.print(e, properties=["fail", "bold"])
            continue

        soup = BeautifulSoup(response.text, "html.parser")

        raw_title = soup.find("div", class_="sku-title")
        if raw_title is None:
            Colors.print(
                f"Unable to find title for {url}.", properties=["fail", "bold"]
            )
            continue
        title = raw_title.get_text().strip()

        in_stock = soup.find(
            "button",
            {"data-button-state": "ADD_TO_CART"},
        )
        if in_stock:
            Colors.warn(f"{title} is already in stock.")

        yield title

    session.close()


def test_urls():
    with loading("Testing URLs"):
        for title in get_url_titles():
            Colors.print(f"Confirmed {title}!", properties=["success"])


def clear_urls():
    _save({})
    Colors.print("Successfully removed all URLs!", properties=["success"])



import glob
import os.path
from pathlib import Path


def calculate_indentation(string):
    return string[: string.index(string.strip())]


def starts_with(string, *substrings, begin=0, end=None):
    starts_with_substring = list(map(string[begin:end].startswith, substrings))
    if any(starts_with_substring):
        return substrings[starts_with_substring.index(True)]
    return False


MSG = '''"""
EASTER EGG: Thank you for reading the source code!
To run the bot with a higher priority level and achieve better performance complete the following.

If using Firefox, complete the following before moving on to the next step:
WINDOWS: Open a Command Prompt window with "Run as administrator" https://www.educative.io/edpresso/how-to-run-cmd-as-an-administrator
MAC: Enter the command `su` in your terminal to gain root privileges. Beware your settings may be different in the root session, but you can always return to a normal session with the `exit` command.

Then regardless of your browser:
Run `3b-bot --fast` in your shell.
"""'''

split_msg = MSG.split("\n")


def indent_msg(indent):
    return "\n".join(
        [indent + msg_line if msg_line.strip() else "" for msg_line in split_msg]
    )


if __name__ == "__main__":
    # Get path of best_buy_bullet_bot directory
    parent_dir = Path(__file__).parents[1]
    search_dir = os.path.join(parent_dir, "best_buy_bullet_bot")

    for filename in glob.iglob(os.path.join(search_dir, "**"), recursive=True):
        # Skip if not a python file
        if not filename.endswith(".py"):
            continue

        with open(filename, "r+") as f:
            code = f.read()
            lowercase_code = code.lower()

            # Skip file if no easter egg comments need to be added
            if "easter egg" not in lowercase_code:
                continue

            lines = code.split("\n")
            lower_lines = lowercase_code.split("\n")

            for idx, line in enumerate(lines):
                line = line.lower().strip()

                # Skip line if the text "easter egg" is not in it
                if "easter egg" not in line:
                    continue

                # This variable means we will delete the following lines until we find a line that ends in the variable
                clear_multiline_string = starts_with(line, "'''", '"""')

                # If the multiline comment starts on the previous line
                if not clear_multiline_string and not starts_with(line, "'", '"', "#"):
                    previous_line = lines[idx - 1]
                    indent = calculate_indentation(previous_line)

                    previous_line = previous_line.strip()
                    clear_multiline_string = starts_with(previous_line, "'''", '"""')

                    if clear_multiline_string:
                        # Delete the previous line
                        lines.pop(idx - 1)
                        idx -= 1
                    else:
                        # Its not a comment, just the text "easter egg" laying around somewhere
                        continue
                else:
                    indent = calculate_indentation(lines[idx])

                if clear_multiline_string:
                    # Delete all subsequent lines until the comment ends
                    while not lines[idx + 1].strip().endswith(clear_multiline_string):
                        lines.pop(idx + 1)
                    lines.pop(idx + 1)

                # Replace the current line with the correct message
                lines.pop(idx)
                lines.insert(idx, indent_msg(indent))

            easter_egg_code = "\n".join(lines)

            # Update the file with the new code
            if easter_egg_code != code:
                f.seek(0)
                f.write(easter_egg_code)
                f.truncate()



import sys

from rich.traceback import install

from best_buy_bullet_bot.command_line import run_command


def main():
    # Stylizes errors and shows more info
    install()

    try:
        run_command()
    except KeyboardInterrupt:
        # This way we don't get the keyboardinterrupt traceback error
        sys.exit(0)


# Running things directly can be useful in development
if __name__ == "__main__":
    main()
